"""Simplified python debuging."""

from .debug import msg, once, msg1, table, file, file1, table_flush_header, table_flush_line, get_line_number_file

__all__ = [
    'msg',
    'once',
    'msg1',
    'table',
    'file',
    'file1',
    'table_flush_header',
    'table_flush_line',
    'get_line_number_file',
]
